From afe5333a37204d6fef809449095ecc8dfd36b733 Mon Sep 17 00:00:00 2001
From: Paul Eggert <eggert@cs.ucla.edu>
Date: Mon, 12 Dec 2016 15:34:22 -0800
Subject: [PROPOSED] Add string interpolation to 'format'

This still needs to be documented.
* src/editfns.c (styled_format): Support interleaved formats and
arguments, so long as each early format contains only one
conversion specificiation at the end, followed by a '%'
continuation marker so that this is a pure extension to Emacs
behavior.  See the thread rooted here:
http://lists.gnu.org/archive/html/emacs-devel/2016-12/msg00493.html
---
 src/editfns.c | 85 +++++++++++++++++++++++++++++++++++++++++------------------
 1 file changed, 60 insertions(+), 25 deletions(-)

diff --git a/src/editfns.c b/src/editfns.c
index 6ea8cba..7cffd8a 100644
--- a/src/editfns.c
+++ b/src/editfns.c
@@ -3968,15 +3968,41 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
     bool_bf intervals : 1;
   } *info;
 
+  ptrdiff_t format_index = 0;
   CHECK_STRING (args[0]);
-  char *format_start = SSDATA (args[0]);
-  ptrdiff_t formatlen = SBYTES (args[0]);
+  char *format_start;
+
+  /* Compute a bound on the maximum number of bytes of any format string.
+
+     Also, try to determine whether the result should be multibyte.
+     This is not always right; sometimes the result needs to be
+     multibyte because of an object that we will pass through prin1,
+     or because a grave accent or apostrophe is requoted, and in that
+     case, we won't know it here.  */
+  ptrdiff_t formatlen_bound = 0;
+  bool definitely_found_all_formats = false;
+  bool multibyte = false;
+  for (ptrdiff_t i = 0; i < nargs; i++)
+    if (STRINGP (args[i]))
+      {
+	if (STRING_MULTIBYTE (args[i]))
+	  multibyte = true;
+	if (!definitely_found_all_formats)
+	  {
+	    formatlen_bound = max (formatlen_bound, SBYTES (args[i]));
+	    definitely_found_all_formats
+	      = (SBYTES (args[i]) == 0
+		 || SREF (args[i], SBYTES (args[i]) - 1) != '%');
+	  }
+	if (multibyte && definitely_found_all_formats)
+	  break;
+      }
 
   /* Allocate the info and discarded tables.  */
   ptrdiff_t alloca_size;
   if (INT_MULTIPLY_WRAPV (nargs, sizeof *info, &alloca_size)
       || INT_ADD_WRAPV (sizeof *info, alloca_size, &alloca_size)
-      || INT_ADD_WRAPV (formatlen, alloca_size, &alloca_size)
+      || INT_ADD_WRAPV (formatlen_bound, alloca_size, &alloca_size)
       || SIZE_MAX < alloca_size)
     memory_full (SIZE_MAX);
   /* info[0] is unused.  Unused elements have -1 for start.  */
@@ -3989,20 +4015,8 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
      It is 2 if byte I was not the first byte of its character.  */
   char *discarded = (char *) &info[nargs + 1];
 
-  /* Try to determine whether the result should be multibyte.
-     This is not always right; sometimes the result needs to be multibyte
-     because of an object that we will pass through prin1.
-     or because a grave accent or apostrophe is requoted,
-     and in that case, we won't know it here.  */
-
   /* True if the format is multibyte.  */
-  bool multibyte_format = STRING_MULTIBYTE (args[0]);
-  /* True if the output should be a multibyte string,
-     which is true if any of the inputs is one.  */
-  bool multibyte = multibyte_format;
-  for (ptrdiff_t i = 1; !multibyte && i < nargs; i++)
-    if (STRINGP (args[i]) && STRING_MULTIBYTE (args[i]))
-      multibyte = true;
+  bool multibyte_format;
 
   int quoting_style = message ? text_quoting_style () : -1;
 
@@ -4013,11 +4027,15 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
   p = buf;
   nchars = 0;
   n = 0;
+  maybe_combine_byte = false;
+
+  /* Jump back to here when finding the next format.  */
+ next_format:
 
   /* Scan the format and store result in BUF.  */
-  format = format_start;
-  end = format + formatlen;
-  maybe_combine_byte = false;
+  format = format_start = SSDATA (args[format_index]);
+  end = format + SBYTES (args[format_index]);
+  multibyte_format = STRING_MULTIBYTE (args[format_index]);
 
   while (format != end)
     {
@@ -4026,21 +4044,23 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
       char *format0 = format;
       char const *convsrc = format;
       unsigned char format_char = *format++;
+      ptrdiff_t used;
 
       /* Bytes needed to represent the output of this conversion.  */
       ptrdiff_t convbytes = 1;
 
       if (format_char == '%')
 	{
-	  /* General format specifications look like
+	  /* General conversion specifications look like
 
-	     '%' [flags] [field-width] [precision] format
+	     '%' [flags] [field-width] [precision] format [continue]
 
 	     where
 
 	     flags ::= [-+0# ]+
 	     field-width ::= [0-9]+
 	     precision ::= '.' [0-9]*
+	     continue ::= '%' (only after first spec and at string end)
 
 	     If a field-width is specified, it specifies to which width
 	     the output should be padded with blanks, if the output
@@ -4048,7 +4068,11 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
 
 	     If precision is specified, it specifies the number of
 	     digits to print after the '.' for floats, or the max.
-	     number of chars to print from a string.  */
+	     number of chars to print from a string.
+
+	     If a format string's first conversion spec is followed by
+	     '%' and then the string end, the argument after the
+	     formatted value continues the format.  */
 
 	  bool minus_flag = false;
 	  bool  plus_flag = false;
@@ -4233,7 +4257,7 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
 		  if (string_intervals (args[n]))
 		    info[n].intervals = arg_intervals = true;
 
-		  continue;
+		  goto next_format_check;
 		}
 	    }
 	  else if (! (conversion == 'c' || conversion == 'd'
@@ -4477,7 +4501,7 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
 		    }
 		  info[n].end = nchars;
 
-		  continue;
+		  goto next_format_check;
 		}
 	    }
 	}
@@ -4538,7 +4562,7 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
 	 character.  CONVBYTES says how much room is needed.  Allocate
 	 enough room (and then some) and do it again.  */
 
-      ptrdiff_t used = p - buf;
+      used = p - buf;
       if (max_bufsize - used < convbytes)
 	string_overflow ();
       bufsize = used + convbytes;
@@ -4561,6 +4585,17 @@ styled_format (ptrdiff_t nargs, Lisp_Object *args, bool message)
       p = buf + used;
       format = format0;
       n = n0;
+      continue;
+
+    next_format_check:
+      if (format + 1 == end && *format == '%'
+	  && n == format_index + 1 && n + 1 < nargs)
+	{
+	  n++;
+	  CHECK_STRING (args[n]);
+	  format_index = n;
+	  goto next_format;
+	}
     }
 
   if (bufsize < p - buf)
-- 
2.7.4

